@@ -10,9 +10,11 @@ Import::
     >>> from s3pathlib import S3Path
 """
 
+from typing import Tuple, List, Iterable, Union, Optional, Any
 import functools
 from datetime import datetime
-from typing import Tuple, List, Iterable, Union, Optional, Any
+from urllib.parse import urlencode
+
 from pathlib_mate import Path
 from boto_session_manager import BotoSesManager, AwsServiceEnum
 
@@ -38,8 +40,11 @@ except:  # pragma: no cover
     raise
 
 from . import utils, exc, validate
+from . import client as better_client
+from .type import TagType, MetadataType
 from .aws import context, Context
 from .iterproxy import IterProxy
+from .marker import warn_deprecate
 
 
 class S3PathIterProxy(IterProxy):
@@ -72,6 +77,9 @@ class S3PathIterProxy(IterProxy):
     def all(self) -> List['S3Path']:
         return super(S3PathIterProxy, self).all()
 
+    def skip(self, k: int) -> 'S3PathIterProxy':
+        return super(S3PathIterProxy, self).skip(k=k)
+
     def filter_by_ext(self, *exts: str) -> 'S3PathIterProxy':
         """
         Filter S3 object by file extension. Case is insensitive.
@@ -368,12 +376,13 @@ class S3Path:
         "_cached_cparts",  # cached comparison parts
         "_hash",  # cached hash value
         "_meta",  # s3 object metadata cache object
+        "_tags",  # s3 object tags cache object
     )
 
     # --------------------------------------------------------------------------
     #               Objective Oriented Programming Implementation
     # --------------------------------------------------------------------------
-    __OOP_IMPLEMENTATION__ = None
+    __S1_OOP_IMPLEMENTATION__ = None
 
     def __new__(
         cls,
@@ -453,6 +462,7 @@ class S3Path:
         self._parts = parts
         self._is_dir = is_dir
         self._meta = None
+        self._tags = None
         if init:
             self._init()
         return self
@@ -583,7 +593,7 @@ class S3Path:
     # --------------------------------------------------------------------------
     #                Method that DOESN't need boto3 API call
     # --------------------------------------------------------------------------
-    __ATTRIBUTE_AND_METHOD__ = None
+    __S2_ATTRIBUTE_AND_METHOD__ = None
 
     def to_dict(self) -> dict:
         """
@@ -673,6 +683,58 @@ class S3Path:
             self._hash = hash(tuple(self._cparts))
             return self._hash
 
+    def __truediv__(
+        self,
+        other: Union[
+            str,
+            'S3Path',
+            List[Union[str, 'S3Path']]
+        ]
+    ) -> 'S3Path':
+        """
+        A syntax sugar. Basically ``S3Path(s3path, part1, part2, ...)``
+        is equal to ``s3path / part1 / part2 / ...`` or
+        ``s3path / [part1, part2]``
+
+        Example::
+
+            >>> S3Path("bucket") / "folder" / "file.txt"
+            S3Path('s3://bucket/folder/file.txt')
+
+            >>> S3Path("bucket") / ["folder", "file.txt"]
+            S3Path('s3://bucket/folder/file.txt')
+
+            # relative path also work
+            >>> S3Path("new-bucket") / (S3Path("bucket/file.txt").relative_to(S3Path("bucket")))
+            S3Path('s3://new-bucket/file.txt')
+
+        .. versionadded:: 1.0.11
+        """
+        if self.is_void():
+            raise TypeError("You cannot do ``VoidS3Path / other``!")
+        if isinstance(other, list):
+            res = self
+            for part in other:
+                res = res / part
+            return res
+        else:
+            if (
+                self.is_relpath()
+                and isinstance(other, S3Path)
+                and other.is_relpath() is False
+            ):
+                raise TypeError("you cannot do ``RelativeS3Path / NonRelativeS3Path``!")
+            return S3Path(self, other)
+
+    def __sub__(self, other: 'S3Path') -> 'S3Path':
+        """
+        A syntax sugar. Basically ``s3path1 - s3path2`` is equal to
+        ``s3path2.relative_to(s3path1)``
+
+        .. versionadded:: 1.0.11
+        """
+        return self.relative_to(other)
+
     def copy(self) -> 'S3Path':
         """
         Create an duplicate copy of S3Path object that logically equals to
@@ -1013,7 +1075,7 @@ class S3Path:
         else:
             return "s3://{}/".format(self._bucket)
 
-    @FilterableProperty
+    @property
     def console_url(self) -> Optional[str]:
         """
         Return an AWS S3 Console url that can inspect the details.
@@ -1027,7 +1089,7 @@ class S3Path:
             console_url = utils.make_s3_console_url(s3_uri=uri)
             return console_url
 
-    @FilterableProperty
+    @property
     def us_gov_cloud_console_url(self) -> Optional[str]:
         """
         Return an AWS US Gov Cloud S3 Console url that can inspect the details.
@@ -1043,6 +1105,39 @@ class S3Path:
             )
             return console_url
 
+    @property
+    def s3_select_console_url(self) -> Optional[str]:
+        """
+        Return an AWS US Gov Cloud S3 Console url that can inspect data with s3 select.
+
+        .. versionadded:: 1.0.12
+        """
+        if self.is_file():
+            return utils.make_s3_select_console_url(
+                bucket=self.bucket,
+                key=self.key,
+                is_us_gov_cloud=False,
+            )
+        else:
+            raise TypeError("you can only do s3 select with an object!")
+
+    @property
+    def s3_select_us_gov_cloud_console_url(self) -> Optional[str]:
+        """
+
+        Return an AWS S3 Console url that can inspect data with s3 select.
+
+        .. versionadded:: 1.0.12
+        """
+        if self.is_file():
+            return utils.make_s3_select_console_url(
+                bucket=self.bucket,
+                key=self.key,
+                is_us_gov_cloud=True,
+            )
+        else:
+            raise TypeError("you can only do s3 select with an object!")
+
     @FilterableProperty
     def arn(self) -> Optional[str]:
         """
@@ -1306,12 +1401,16 @@ class S3Path:
 
     def __repr__(self):
         classname = self.__class__.__name__
-        if self.is_relpath():
+
+        if self.is_void():
+            return "S3VoidPath()"
+
+        elif self.is_relpath():
             key = self.key
             if len(key):
-                return "{}('{}')".format(classname, key)
+                return f"S3RelPath({key!r})"
             else:
-                return "{}()".format(classname)
+                return "S3RelPath()"
         else:
             uri = self.uri
             return "{}('{}')".format(classname, uri)
@@ -1323,10 +1422,14 @@ class S3Path:
         """
         Return the relative path to another path. If the operation
         is not possible (because this is not a sub path of the other path),
-        raise ValueError.
+        raise ``ValueError``.
+
+        ``-`` is a syntax sugar for ``relative_to``. See more information at
+        :meth:`~S3Path.__sub__`.
 
         The relative path usually works with :meth:`join_path` to form a new
-        path.
+        path. Or you can use the ``/`` syntax sugar as well. See more
+        information at :meth:`~S3Path.__truediv__`.
 
         Examples::
 
@@ -1341,7 +1444,7 @@ class S3Path:
 
         :param other: other :class:`S3Path` instance.
 
-        :return: an relative path object, which is a special version of S3Path
+        :return: a relative path object, which is a special version of S3Path
 
         .. versionadded:: 1.0.1
         """
@@ -1406,6 +1509,11 @@ class S3Path:
 
         .. versionadded:: 1.0.1
         """
+        warn_deprecate(
+            func_name="S3Path.join_path",
+            version="2.1.1",
+            message="S3Path.joinpath",
+        )
         args = [self, ]
         for relp in others:
             if relp.is_relpath() is False:
@@ -1417,6 +1525,51 @@ class S3Path:
             args.append(relp)
         return self._from_parts(args)
 
+    def joinpath(
+        self,
+        *other: Union[str, 'S3Path']
+    ) -> 'S3Path':
+        """
+        Join with other relative path or string parts.
+
+        Example::
+
+            # join with string parts
+            >>> p = S3Path("bucket")
+            >>> p.joinpath("folder", "file.txt")
+            S3Path('s3://bucket/folder/file.txt')
+
+            # join ith relative path or string parts
+            >>> p = S3Path("bucket")
+            >>> relpath = S3Path("my-bucket", "data", "folder/").relative_to(S3Path("my-bucket", "data"))
+            >>> p.joinpath("data", relpath, "file.txt")
+            S3Path('s3://bucket/data/folder/file.txt')
+
+        :param others: many string or relative path
+
+        .. versionadded:: 1.1.1
+        """
+        args = [self, ]
+        for part in other:
+            if isinstance(part, str):
+                args.append(part)
+            elif isinstance(part, S3Path):
+                if part.is_relpath() is False:
+                    msg = (
+                        "you can only join with string part or relative path! "
+                        "{} is not a relative path"
+                    ).format(part)
+                    raise TypeError(msg)
+                else:
+                    args.append(part)
+            else:
+                msg = (
+                    "you can only join with string part or relative path! "
+                    "{} is not a relative path"
+                ).format(part)
+                raise TypeError(msg)
+        return self._from_parts(args)
+
     def ensure_object(self) -> None:
         """
         A validator method that ensure it represents a S3 object.
@@ -1449,7 +1602,7 @@ class S3Path:
     # --------------------------------------------------------------------------
     #                   Method that need boto3 API call
     # --------------------------------------------------------------------------
-    __STATELESS_API__ = None
+    __S3_STATELESS_API__ = None
 
     def clear_cache(self) -> None:
         """
@@ -1560,6 +1713,54 @@ class S3Path:
         """
         return self._get_meta_value(key="Metadata", default=dict())
 
+    def update_metadata(self, metadata: dict):  # pragma: no cover
+        raise NotImplementedError(
+            "You CANNOT only update metadata without changing the content of the "
+            "object! You can only do full replace ment via the .write_text() or "
+            ".write_bytes() API. This method will NEVER be implemented!"
+        )
+
+    def get_tags(self, bsm: Optional['BotoSesManager'] = None) -> TagType:
+        """
+        Get s3 object tags in key value pairs dict.
+
+        :return: the s3 object tags in string key value pairs dict.
+
+        .. versionadded:: 1.1.1
+        """
+        self.ensure_object()
+        s3_client = _resolve_s3_client(context, bsm)
+        return better_client.get_object_tagging(s3_client, self.bucket, self.key)
+
+    def put_tags(self, tags: TagType, bsm: Optional['BotoSesManager'] = None) -> TagType:
+        """
+        Do full replacement of s3 object tags.
+
+        :param tags: the s3 object tags in string key value pairs dict.
+
+        :return: the s3 object tags in string key value pairs dict.
+
+        .. versionadded:: 1.1.1
+        """
+        self.ensure_object()
+        s3_client = _resolve_s3_client(context, bsm)
+        better_client.put_object_tagging(s3_client, self.bucket, self.key, tags)
+        return tags
+
+    def update_tags(self, tags: TagType, bsm: Optional['BotoSesManager'] = None) -> TagType:
+        """
+        Do partial updates of s3 object tags.
+
+        :param tags: the s3 object tags in string key value pairs dict.
+
+        :return: the latest, merged object tags in string key value pairs dict.
+
+        .. versionadded:: 1.1.1
+        """
+        self.ensure_object()
+        s3_client = _resolve_s3_client(context, bsm)
+        return better_client.update_object_tagging(s3_client, self.bucket, self.key, tags)
+
     def exists(self, bsm: Optional['BotoSesManager'] = None) -> bool:
         """
         - For S3 bucket: check if the bucket exists. If you don't have the
@@ -1619,7 +1820,7 @@ class S3Path:
     # --------------------------------------------------------------------------
     #            Method that change the state of S3 bucket / objects
     # --------------------------------------------------------------------------
-    __STATEFUL_API__ = None
+    __S4_STATEFUL_API__ = None
 
     def upload_file(
         self,
@@ -2039,15 +2240,16 @@ class S3Path:
 
     def open(
         self,
-        mode="r",
-        buffering=-1,
-        encoding=None,
-        errors=None,
-        newline=None,
+        mode: Optional[str] = "r",
+        buffering: Optional[int] = -1,
+        encoding: Optional[str] = None,
+        errors: Optional[str] = None,
+        newline: Optional[str] = None,
         closefd=True,
         opener=None,
-        ignore_ext=False,
-        compression=None,
+        ignore_ext: bool = False,
+        compression: Optional[str] = None,
+        transport_params: Optional[dict] = None,
         bsm: Optional['BotoSesManager'] = None,
     ):
         """
@@ -2058,6 +2260,9 @@ class S3Path:
         See https://github.com/RaRe-Technologies/smart_open for more info.
         """
         s3_client = _resolve_s3_client(context, bsm)
+        if transport_params is None:
+            transport_params = dict()
+        transport_params["client"] = s3_client
         kwargs = dict(
             uri=self.uri,
             mode=mode,
@@ -2067,7 +2272,7 @@ class S3Path:
             newline=newline,
             closefd=closefd,
             opener=opener,
-            transport_params={"client": s3_client}
+            transport_params=transport_params,
         )
         if smart_open_version_major < 6:  # pragma: no cover
             kwargs["ignore_ext"] = ignore_ext
@@ -2099,29 +2304,86 @@ class S3Path:
         data: str,
         encoding="utf-8",
         errors=None,
-        newline=None,
+        metadata: Optional[MetadataType] = None,
+        tags: Optional[TagType] = None,
         bsm: Optional['BotoSesManager'] = None,
     ):
-        with self.open(
-            mode="w",
-            encoding=encoding,
-            errors=errors,
-            newline=newline,
-            bsm=bsm
-        ) as f:
-            f.write(data)
+        """
+        Write text to s3 object. A simple wrapper around
+        `s3_client.put_object <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Bucket.put_object>`_
 
-    def write_bytes(self, data: bytes, bsm: Optional['BotoSesManager'] = None):
-        with self.open(mode="wb", bsm=bsm) as f:
-            f.write(data)
+        :param data: the text you want to write.
+        :param encoding: how do you want to encode text?
+        :param errors: how do you want to handle encode error? can be 'strict',
+            ``ignore``, ``replace``, ``xmlcharrefreplace``, ``backslashreplace``.
+            see more details `here <https://docs.python.org/3/library/stdtypes.html#str.encode>`_.
+        :param metadata: the s3 object metadata in string key value pair dict.
+        :param tags: the s3 object tags in string key value pair dict.
+        
+        .. versionadded:: 1.0.3
+        
+        .. versionchanged:: 1.1.1
+        """
+        s3_client = _resolve_s3_client(context, bsm)
+        if errors:
+            body = data.encode(encoding, errors=errors)
+        else:
+            body = data.encode(encoding)
+        response = better_client.put_object(
+            s3_client=s3_client,
+            bucket=self.bucket,
+            key=self.key,
+            body=body,
+            metadata=metadata,
+            tags=tags,
+        )
+        self._tags = tags
+        return response
+
+    def write_bytes(
+        self,
+        data: bytes,
+        metadata: Optional[dict] = None,
+        tags: Optional[TagType] = None,
+        bsm: Optional['BotoSesManager'] = None,
+    ):
+        """
+        Write binary data to s3 object. A simple wrapper around
+        `s3_client.put_object <https://boto3.amazonaws.com/v1/documentation/api/latest/reference/services/s3.html#S3.Bucket.put_object>`_
+
+        :param data: the text you want to write.
+        :param metadata: the s3 object metadata in string key value pair dict.
+        :param tags: the s3 object tags in string key value pair dict.
+
+        .. versionadded:: 1.0.3
+
+        .. versionchanged:: 1.1.1
+        """
+        s3_client = _resolve_s3_client(context, bsm)
+        response = better_client.put_object(
+            s3_client=s3_client,
+            bucket=self.bucket,
+            key=self.key,
+            body=data,
+            metadata=metadata,
+            tags=tags,
+        )
+        self._tags = tags
+        return response
 
     def touch(
         self,
         exist_ok: bool = True,
         bsm: Optional['BotoSesManager'] = None,
     ):  # pragma: no cover
-        if not self.is_file():
-            raise ValueError
+        """
+        Create an empty S3 object at the S3 location if the S3 object not exists.
+        Do nothing if already exists.
+
+        :param exist_ok: if True, it won't raise error when the S3 object
+            already exists.
+        """
+        self.ensure_object()
 
         if self.exists(bsm=bsm):
             if exist_ok:
@@ -2162,3 +2424,29 @@ class S3Path:
             for p in self.parents:
                 if p.is_bucket() is False:
                     p.mkdir(exist_ok=True, parents=False, bsm=bsm)
+
+    # --------------------------------------------------------------------------
+    #                         S3 Object Tagging
+    # --------------------------------------------------------------------------
+    __S5_TAGGING__ = None
+
+    def get_tagging(self, *args, **kwargs) -> dict:
+        raise NotImplementedError
+
+    def put_tagging(self, *args, **kwargs) -> dict:
+        raise NotImplementedError
+
+    def delete_tagging(self, *args, **kwargs) -> dict:
+        raise NotImplementedError
+
+    def get_acl(self, *args, **kwargs) -> dict:
+        raise NotImplementedError
+
+    def put_acl(self, *args, **kwargs) -> dict:
+        raise NotImplementedError
+
+    def get_legal_hold(self, *args, **kwargs) -> dict:
+        raise NotImplementedError
+
+    def put_legal_hold(self, *args, **kwargs) -> dict:
+        raise NotImplementedError
