@@ -4,6 +4,7 @@
 Exception creator and helpers, argument validators, and more.
 """
 
+
 def ensure_one_and_only_one_not_none(**kwargs) -> None:
     """
     Ensure only exact one of the keyword argument is not None.
@@ -16,6 +17,7 @@ def ensure_one_and_only_one_not_none(**kwargs) -> None:
             f"{list(kwargs)} can be not None!"
         )
 
+
 def ensure_all_none(**kwargs) -> None:
     """
     Ensure all of the keyword argument is None.
@@ -25,4 +27,16 @@ def ensure_all_none(**kwargs) -> None:
     if sum([v is not None for _, v in kwargs.items()]) != 0:
         raise ValueError(
             f"arguments from {list(kwargs)} has to be all None!"
-        )
\ No newline at end of file
+        )
+
+
+class S3BucketNotExist(Exception):
+    pass
+
+
+class S3ObjectNotExist(Exception):
+    pass
+
+
+class S3PermissionDenied(Exception):
+    pass
